# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the core `Tab` object.

The basic Tab allows for simple embedding of arbitrary text inside a
standardised HTML interface. Most real-world applications will use a
sub-class of `Tab` to create more complex HTML output.

The `Tab` class comes in three flavours:

- 'static', no specific GPS time reference
- 'interval', displaying data in a given GPS [start, stop) interval
- 'event', displaying data around a specific central GPS time

The flavour is dynamically set when each instance is created based on the
`mode` keyword, or the presence of `span`, `start and `end`, or `gpstime`
keyword arguments.
"""

import os
import re
from collections import OrderedDict
from configparser import NoOptionError

from MarkupPy import markup

from gwpy.time import (from_gps, to_gps)
from gwpy.segments import Segment

from gwdetchar.io import html as gwhtml

from .. import html
from ..mode import (Mode, get_mode, get_base)
from ..utils import (re_quote, re_cchar)
from .registry import (get_tab, register_tab)
from .._version import get_versions

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['BaseTab', 'Tab', 'TabList']


# -- BaseTab ------------------------------------------------------------------
# this object defines the basic object from which all three flavours inherit

class BaseTab(object):
    """The core `Tab` object, defining basic functionality
    """
    def __init__(self, name, index=None, shortname=None, parent=None,
                 children=list(), group=None, notes=None, overlay=None,
                 path=os.curdir, mode=None, hidden=False):
        # mode
        self.mode = mode
        # names
        self.name = name
        self.shortname = shortname
        # structure
        self._children = []
        self.parent = parent
        self.children = children
        self.group = group
        self.notes = notes
        self.overlay = overlay
        # HTML format
        self.path = path
        self.index = index
        self.page = None
        self.hidden = hidden

    # -- properties -----------------------------

    @property
    def name(self):
        """Full name for this `Tab`

        :type: `str`
        """
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def shortname(self):
        """Short name for this tab

        This will be displayed in the navigation bar.

        :type: `str`
        """
        return self._shortname or self.name

    @shortname.setter
    def shortname(self, name):
        self._shortname = name

    @property
    def parent(self):
        """Short name of the parent page for this `Tab`

        A given tab can either be a parent for a set of child tabs, or can
        have a parent, it cannot be both. In this system, the `parent`
        attribute defines the heading under which this tab will be linked
        in the HTML navigation bar.

        :type: `str`
        """
        return self._parent

    @parent.setter
    def parent(self, p):
        if p is None:
            del self.parent
        else:
            self.set_parent(p)

    @parent.deleter
    def parent(self):
        self._parent = None

    def set_parent(self, p):
        """Set the parent `Tab` for this tab

        Parameters
        ----------
        p : `Tab`
            the parent tab for this one
        """
        if p and self._children:
            raise ValueError("A tab cannot have both a parent, and a "
                             "set of children.")
        if isinstance(p, BaseTab) and self not in p.children:
            p.add_child(self)
        else:
            p = ParentTab(p, [self], mode=self.mode)
        self._parent = p

    @property
    def children(self):
        """List of child tabs for this `Tab`

        If this tab is given children, it cannot also have a parent, as it
        will define its own dropdown menu in the HTML navigation bar, linking
        to itself and its children.

        :type: `list` of `tabs <Tab>`
        """
        return self._children

    @children.setter
    def children(self, clist):
        if self._parent and clist:
            raise ValueError("A Tab cannot have both a parent, and a "
                             "set of children.")
        self._children = list(clist)

    @property
    def index(self):
        """The HTML path (relative to the `~Tab.path`) for this tab
        """
        if not self._index:
            if self.shortname.lower() == 'summary':
                p = ''
            else:
                p = re_cchar.sub('_', self.shortname.strip('_')).lower()
            tab_ = self
            while tab_.parent:
                p = os.path.join(re_cchar.sub(
                        '_', tab_.parent.shortname.strip('_')).lower(), p)
                tab_ = tab_.parent
            self._index = os.path.normpath(os.path.join(
                self.path, p, 'index.html'))
        return self._index

    @index.setter
    def index(self, p):
        self._index = p

    @index.deleter
    def index(self):
        self._index = None

    @property
    def href(self):
        """HTML href (relative to the `~Tab.path`) for this tab

        This attribute is just a convenience to clean-up the
        `~Tab.index` for a given tab, by removing index.htmls.
        hierarchy.

        :type: `str`
        """
        if os.path.basename(self.index) in ('index.html', 'index.php'):
            return os.path.split(self.index)[0] + os.path.sep
        elif self.index:
            return self.index
        else:
            return ''

    @property
    def title(self):
        """Page title for this tab
        """
        if self.parent:
            title = self.name.strip('_')
            tab_ = self
            while tab_.parent:
                title = '%s/%s' % (tab_.parent.name.strip('_'), title)
                tab_ = tab_.parent
            return title
        else:
            return self.name

    @property
    def shorttitle(self):
        """Page title for this tab
        """
        if self.parent:
            title = self.shortname.strip('_')
            tab_ = self
            while tab_.parent:
                title = '%s/%s' % (tab_.parent.shortname.strip('_'), title)
                tab_ = tab_.parent
            return title
        else:
            return self.shortname.strip('_')

    @property
    def group(self):
        """Dropdown group for this `Tab` in the navigation bar

        :type: `str`
        """
        return self._group

    @group.setter
    def group(self, gp):
        if gp is None:
            self._group = None
        else:
            self._group = str(gp)

    @property
    def notes(self):
        """Release notes for this `Tab`
        """
        return self._notes

    @notes.setter
    def notes(self, n):
        if n is None:
            self._notes = None
        else:
            self._notes = n

    @property
    def overlay(self):
        """Boolean switch to enable plot overlay for this `Tab`
        """
        return self._overlay

    @overlay.setter
    def overlay(self, ovl):
        self._overlay = ovl

    @property
    def mode(self):
        """The date-time mode of this tab.

        :type: `int`

        See Also
        --------
        gwsumm.mode : for details on the modes
        """
        return self._mode

    @mode.setter
    def mode(self, m):
        self._mode = get_mode(m)

    @mode.deleter
    def mode(self):
        self._mode = get_mode(0)

    # -- Tab instance methods -------------------

    def add_child(self, tab):
        """Add a child to this `SummaryTab`

        Parameters
        ----------
        tab : `SummaryTab`
            child tab to record
        """
        self.children.append(tab)

    def get_child(self, name):
        """Find a child tab of this `SummaryTab` by name

        Parameters
        ----------
        name : `str`
            string identifier of child tab to use in search

        Returns
        -------
        child : `SummaryTab`
            the child tab found by name

        Raises
        ------
        RuntimeError
            if no child tab can be found matching the given ``name``
        """
        names = [c.name for c in self.children]
        try:
            idx = names.index(name)
        except ValueError:
            raise RuntimeError("This tab has no child named '%s'." % name)
        else:
            return self.children[idx]

    # -- Tab configuration parser ---------------

    @classmethod
    def from_ini(cls, cp, section, *args, **kwargs):
        """Define a new tab from a `~gwsumm.config.GWConfigParser`

        Parameters
        ----------
        cp : `~gwsumm.config.GWConfigParser`
            customised configuration parser containing given section

        section : `str`
            name of section to parse

        *args, **kwargs
            other positional and keyword arguments to pass to the class
            constructor (`__init__`)

        Returns
        -------
        tab : `Tab`
            a new tab defined from the configuration

        Notes
        -----
        This method parses the following configuration options

        .. autosummary::

           ~Tab.name
           ~Tab.shortname
           ~Tab.parent
           ~Tab.group
           ~Tab.notes
           ~Tab.overlay
           ~Tab.index

        Sub-classes should parse their own configuration values and then pass
        these as ``*args`` and ``**kwargs`` to this method via `super`:

        .. code-block:: python

           class MyTab(Tab):
               [...]
               def from_ini(cls, cp, section)
                   \"\"\"Define a new `MyTab`.
                   \"\"\"
                   foo = cp.get(section, 'foo')
                   bar = cp.get(section, 'bar')
                   return super(MyTab, cls).from_ini(cp, section, foo, bar=bar)
        """
        # get tab name
        try:
            # name given explicitly
            name = re_quote.sub('', cp.get(section, 'name'))
        except NoOptionError:
            # otherwise strip 'tab-' from section name
            name = section[4:]
        try:
            kwargs.setdefault('shortname',
                              re_quote.sub('', cp.get(section, 'shortname')))
        except NoOptionError:
            pass
        # get parent:
        #     if parent is not given, this assumes a top-level tab
        try:
            kwargs.setdefault('parent',
                              re_quote.sub('', cp.get(section, 'parent')))
        except NoOptionError:
            pass
        else:
            if kwargs['parent'] == 'None':
                kwargs['parent'] = None
        # get group
        try:
            kwargs.setdefault('group', cp.get(section, 'group'))
        except NoOptionError:
            pass
        # get HTML file
        try:
            kwargs.setdefault('index', cp.get(section, 'index'))
        except NoOptionError:
            pass
        # get hidden param
        try:
            hidden = cp.get(section, 'hidden')
        except NoOptionError:
            hidden = False
        else:
            if hidden is None:
                hidden = True
            else:
                hidden = bool(hidden.title())
        kwargs.setdefault('hidden', hidden)

        # get release notes
        try:
            kwargs.setdefault('notes', cp.get(section, 'notes'))
        except NoOptionError:
            pass

        # determine whether plot overlay is requested
        try:
            overlay = cp.get(section, 'overlay')
        except NoOptionError:
            overlay = True
        else:
            if overlay is None:
                overlay = True
            else:
                overlay = bool(overlay.title())
        kwargs.setdefault('overlay', overlay)

        # get mode and times if required
        try:
            kwargs['mode']
        except KeyError:
            try:
                kwargs['mode'] = get_mode(cp.get(section, 'mode'))
            except NoOptionError:
                kwargs['mode'] = get_mode()
        if isinstance(kwargs['mode'], str):
            kwargs['mode'] = get_mode(kwargs['mode'])
        if kwargs['mode'] >= Mode.gps:
            try:
                kwargs['start']
            except KeyError:
                kwargs['start'] = cp.getint(section, 'gps-start-time')
            try:
                kwargs['end']
            except KeyError:
                kwargs['end'] = cp.getint(section, 'gps-end-time')
        elif kwargs['mode'] == Mode.event:
            try:
                kwargs['gpstime']
            except KeyError:
                kwargs['gpstime'] = cp.getfloat(section, 'gpstime')
            try:
                kwargs['duration']
            except KeyError:
                kwargs['duration'] = cp.getfloat(section, 'duration')

        return cls(name, *args, **kwargs)

    # -- HTML operations ------------------------
    # the following related HTML operations are defined here
    #
    # - navbar: create navbar
    # - banner: create header
    # - content: create main content
    #
    # The `Tab.write_html` method pulls all of these things together and
    # is the primary user-facing HTML method

    def html_banner(self, title=None, subtitle=None):
        """Build the HTML headline banner for this tab.

        Parameters
        ----------
        title : `str`
            title for this page

        subtitle : `str`
            sub-title for this page

        Returns
        -------
        banner : `~MarkupPy.markup.page`
            formatter markup page for the banner
        """
        # work title as Parent Name/Tab Name
        if title is None and subtitle is None:
            title = self.title.replace('/', ' : ', 1)
        return html.banner(title, subtitle=subtitle)

    def html_navbar(self, help_=None, calendar=[], tabs=list(),
                    ifo=None, ifomap=dict(), **kwargs):
        """Build the navigation bar for this tab.

        Parameters
        ----------
        help_ : `str`, `~MarkupPy.markup.page`
            content to place on the upper-right side of the navbar

        calendar : `list`, optional
            datepicker calendar objects for navigation

        tabs : `list`, optional
            list of parent tabs (each with a list of children) to include
            in the navigation bar.

        ifo : `str`, optional
            prefix for this IFO.

        ifomap : `dict`, optional
            `dict` of (ifo, {base url}) pairs to map to summary pages for
            other IFOs.

        **kwargs
            other keyword arguments to pass to :meth:`gwsumm.html.navbar`

        Returns
        -------
        page : `~MarkupPy.markup.page`
            a markup page containing the navigation bar.
        """
        class_ = 'navbar fixed-top navbar-expand-md shadow-sm'
        # build interferometer cross-links
        if ifo is not None:
            brand_ = html.base_map_dropdown(ifo, id_='ifos', bases=ifomap)
            class_ += ' navbar-%s' % ifo.lower()
        else:
            brand_ = markup.page()
        # build HTML brand
        if help_:
            brand_ = (brand_, help_)
        # build tabs and calendar
        tabs = self._html_navbar_links(tabs)
        if calendar:
            tabs = list(calendar) + tabs
        # combine and return
        return gwhtml.navbar(tabs, class_=class_, brand=brand_, **kwargs)

    def _html_navbar_links(self, tabs):
        """Construct the ordered list of tabs to write into the navbar

        Parameters
        ----------
        tabs : `list`
            a list of `Tabs <gwsumm.tabs.Tab`, some of which may contain
            `~Tab.children`.

        Returns
        -------
        links : `str`
            a structured list of navigation menu entries. Each element will
            be a (`name`, `entries`) tuple defining the heading and dropdown
            entries for a single dropdown menu. The `entries` list will again
            be a list of (`name`, (`link`|`entries`)) tuples, each defining
            a name and link for a given entry in a single dropdown menu,
            or a set of (`name`, `link`) tuples for a group in a dropdown
            menu.
        """
        # build navbar links
        navlinks = []
        tabs = TabList(tabs).get_hierarchy()
        for tab in tabs:
            if tab.hidden:
                continue
            children = [t for t in tab.children if not t.hidden]
            if len(children):
                navlinks.append([tab.shortname.strip('_'), []])
                links = []
                active = None
                # build groups
                groups = set([t.group for t in children])
                groups = dict((g, [t for t in children if t.group == g])
                              for g in groups)
                nogroup = sorted(
                    groups.pop(None, []),
                    key=(lambda c:
                         c.shortname.lower() in ['summary', 'overview'] and
                         c.shortname.upper() or c.shortname.lower()))
                for child in nogroup:
                    links.append((child.shortname.strip('_'), child.href))
                    if child == self:
                        active = len(links) - 1
                for group in sorted(list(groups)):
                    # sort group by name
                    re_group = re.compile(
                        r'(\A{0}\s|\s{0}\Z)'.format(group.strip('_')),
                        re.I,
                    )
                    names = [re_group.sub('', t.shortname)
                             for t in groups[group]]
                    groups[group] = list(zip(*sorted(
                        list(zip(groups[group], names)),
                        key=(lambda x:
                             x[1].lower() in ['summary', 'overview'] and
                             ' %s' % x[1].upper() or x[1].lower()),
                    )))[0]
                    # build link sets
                    links.append((group.strip('_'), []))
                    for i, child in enumerate(groups[group]):
                        name = re_group.sub('', child.shortname.strip('_'))
                        links[-1][1].append((name, child.href))
                        if child == self:
                            active = [len(links) - 1, i]
                if (children[0].shortname == 'Summary' and
                        not children[0].group and len(children)) > 1:
                    links.insert(1, None)
                    if active and isinstance(active, int) and active > 0:
                        active += 1
                    elif active and isinstance(active, list) and active[0] > 0:
                        active[0] += 1
                navlinks[-1][1].extend(links)
                navlinks[-1].append(active)
            else:
                navlinks.append((tab.shortname.strip('_'), tab.href))
        return navlinks

    @staticmethod
    def html_content(content):
        """Build the #main div for this tab.

        Parameters
        ----------
        content : `str`, `~MarkupPy.markup.page`
            HTML content to be wrapped

        Returns
        -------
        #main : `~MarkupPy.markup.page`
            A new `page` with the input content wrapped as
        """
        page = markup.page()
        page.div(id_='main')
        page.div(str(content), id_='content')
        page.div.close()
        return page

    def write_html(self, maincontent, title=None, subtitle=None, tabs=list(),
                   ifo=None, ifomap=dict(), help_=None, base=None, css=None,
                   js=None, about=None, footer=None, issues=True, **inargs):
        """Write the HTML page for this `Tab`.

        Parameters
        ----------
        maincontent : `str`, `~MarkupPy.markup.page`
            simple string content, or a structured `page` of markup to
            embed as the content of the #main div.

        title : `str`, optional, default: {parent.name}
            level 1 heading for this `Tab`.

        subtitle : `str`, optional, default: {self.name}
            level 2 heading for this `Tab`.

        tabs: `list`, optional
            list of top-level tabs (with children) to populate navbar

        ifo : `str`, optional
            prefix for this IFO.

        ifomap : `dict`, optional
            `dict` of (ifo, {base url}) pairs to map to summary pages for
            other IFOs.

        help_ : `str`, `~MarkupPy.markup.page`, optional
            non-menu content for navigation bar

        css : `list`, optional
            list of resolvable URLs for CSS files. See `gwsumm.html.CSS` for
            the default list.

        js : `list`, optional
            list of resolvable URLs for javascript files. See
            `gwumm.html.JS` for the default list.

        about : `str`, optional
            href for the 'About' page

        footer : `str`, `~MarkupPy.markup.page`
            external link, if applicable (linked from an icon in the footer)

        issues : `bool` or `str`, default: `True`
            print link to github.com issue tracker for this package

        **inargs
            other keyword arguments to pass to the
            :meth:`~Tab.build_inner_html` method
        """
        # setup directories
        outdir = os.path.split(self.index)[0]
        if outdir and not os.path.isdir(outdir):
            os.makedirs(outdir)

        # get default style and scripts
        if css is None:
            css = list(html.get_css().values())
        if js is None:
            js = list(html.get_js().values())

        # find relative base path
        if base is None:
            n = len(self.index.split(os.path.sep)) - 1
            base = os.sep.join([os.pardir] * n)
        if not base.endswith('/'):
            base += '/'

        # set default title
        if title is None:
            title = self.shorttitle.replace('/', ' | ')

        # construct navigation
        if tabs:
            navbar = str(self.html_navbar(ifo=ifo, ifomap=ifomap,
                                          tabs=tabs, help_=help_))
        else:
            navbar = None

        # initialize page
        self.page = gwhtml.new_bootstrap_page(
            title=title, base=base, path=self.path,
            css=css, script=js, navbar=navbar)

        # add banner
        self.page.add(str(self.html_banner(
            title=title.replace('|', ':'), subtitle=subtitle)))

        # add help button
        if self.notes is not None:
            self.page.add(str(html.dialog_box(
                self.notes, title='Help', id_='help',
                btntxt=markup.oneliner.i('', class_='fas fa-question'))))

        # add overlay button
        if self.overlay:
            self.page.add(str(html.overlay_canvas()))

        # add #main content
        self.page.add(str(self.html_content(maincontent)))

        # format custom footer
        version = get_versions()['version']
        commit = get_versions()['full-revisionid']
        url = 'https://github.com/gwpy/gwsumm/tree/{}'.format(commit)

        # close page and write
        gwhtml.close_page(self.page, self.index, about=about,
                          link=('gwsumm-%s' % version, url, 'GitHub'),
                          issues=issues, external=footer)
        return


# -- Mixins -------------------------------------------------------------------
#
# All actual `Tab` objects will come in three favours:
#
#     - `StaticTab` - no GPS associations
#     - `IntervalTab` - associated with GPS start and stop time
#     - `EventTab` - associated with central GPS time (and duration)


class StaticTab(BaseTab):
    """Simple `Tab` with no GPS association
    """
    pass


class GpsTab(BaseTab):
    """Stub for GPS-related tabs
    """
    @property
    def span(self):
        """The GPS [start, end) span of this tab.

        :type: `~gwpy.segments.Segment`
        """
        return self._span

    @span.setter
    def span(self, seg):
        if seg:
            self._span = Segment(*map(to_gps, seg))
        else:
            self._span = None

    @property
    def start(self):
        """The GPS start time of this tab.

        :type: `float`
        """
        try:
            return self.span[0]
        except TypeError:
            return None

    @property
    def end(self):
        """The GPS end time of this tab.

        :type: `float`
        """
        try:
            return self.span[1]
        except TypeError:
            return None


class IntervalTab(GpsTab):
    """`Tab` defined within a GPS [start, end) interval
    """
    def __init__(self, *args, **kwargs):
        try:
            span = kwargs.pop('span')
        except KeyError:
            try:
                start = kwargs.pop('start')
                end = kwargs.pop('end')
            except KeyError:
                mode = get_mode(kwargs.get('mode')).name
                raise TypeError("%s() in %r mode needs keyword argument 'span'"
                                " or both 'start' and 'end'"
                                % (type(self).__name__, mode))
            else:
                span = (start, end)
        self.span = span
        super(IntervalTab, self).__init__(*args, **kwargs)

    def html_calendar(self):
        """Build the datepicker calendar for this tab.

        Notes
        -----
        The datetime for the calendar is taken from this tab's `~GpsTab.span`
        """
        date = from_gps(self.start)
        # double-check path matches what is required from custom datepicker
        try:
            requiredpath = get_base(date, mode=self.mode)
        except ValueError:
            return ['%d-%d' % (self.start, self.end)]
        if requiredpath not in self.path:
            raise RuntimeError("Tab path %r inconsistent with required "
                               "format including %r for archive calendar"
                               % (self.path, requiredpath))
        # format calendar
        return html.calendar(date, mode=self.mode)

    def html_navbar(self, help_=None, calendar=True, **kwargs):
        """Build the navigation bar for this `Tab`.

        The navigation bar will consist of a switch for this page linked
        to other interferometer servers, followed by the navbar brand,
        then the full dropdown-based navigation menus configured for the
        given ``tabs`` and their descendents.

        Parameters
        ----------
        help_ : `str`, `~MarkupPy.markup.page`
            content for upper-right of navbar

        ifo : `str`, optional
            prefix for this IFO.

        ifomap : `dict`, optional
            `dict` of (ifo, {base url}) pairs to map to summary pages for
            other IFOs.

        tabs : `list`, optional
            list of parent tabs (each with a list of children) to include
            in the navigation bar.

        Returns
        -------
        page : `~MarkupPy.markup.page`
            a markup page containing the navigation bar.
        """
        # add calendar
        calendar = calendar and self.html_calendar()
        # combine and return
        return super(IntervalTab, self).html_navbar(
            help_=help_, calendar=calendar, **kwargs)


class EventTab(GpsTab):
    """`Tab` defined around a central GPS time
    """
    @property
    def gpstime(self):
        """Central GPS time of this tab

        :type: `~gwpy.time.LIGOTimeGPS`
        """
        return self._gpstime

    @gpstime.setter
    def gpstime(self, t):
        self._gpstime = to_gps(t)

    @property
    def datetime(self):
        return from_gps(self.gpstime)

    @property
    def duration(self):
        """Time duration of this tab, centred on the `gpstime`

        :type: `float`
        """
        return abs(self.span)

    @duration.setter
    def duration(self, d):
        d2 = d/2.
        self.span = (self.gpstime - d2, self.gpstime + d2)

    # -- EventTab methods -----------------------

    def __init__(self, *args, **kwargs):
        # parse gpstime and duration
        try:
            gpstime = kwargs.pop('gpstime')
        except KeyError:
            mode = get_mode(kwargs['mode']).name
            raise TypeError("%s() in %r mode needs keyword argument 'gpstime'"
                            % (type(self).__name__, mode))
        duration = kwargs.pop('duration', 200)
        self.gpstime = gpstime
        self.duration = duration
        # create tab and assign properties
        super(EventTab, self).__init__(*args, **kwargs)

    def html_navbar(self, calendar=[], **kwargs):
        if not calendar:
            calendar = str(self.gpstime)
        super(EventTab, self).html_navbar(calendar=calendar, **kwargs)
    html_navbar.__doc__ = GpsTab.html_navbar.__doc__


class _MetaTab(type):
    """Metaclass for creating tabs of the right flavour

    The `MetaTab.__call__` method will get executed whenever a subclass of
    `Tab` is created. It works out the 'mode' of the new tab and dynamically
    sets the parent class for `Tab` accordingly.
    """
    def __call__(cls, *args, **kwargs):
        """Parse the `mode` kwarg for the Tab and add the right flavour
        """
        # parse default mode based on other kwargs
        try:
            mode = kwargs['mode']
        except KeyError:
            if 'gpstime' in kwargs:
                mode = 'EVENT'
            else:
                mode = None
        mode = get_mode(mode)
        # parse regular Tab (don't add a mixin)
        if mode == Mode.static:
            kwargs.pop('mode', None)
            base = StaticTab
        # parse event Tab
        elif mode == Mode.event:
            kwargs['mode'] = mode
            base = EventTab
        # parse interval Tab
        else:
            kwargs['mode'] = mode
            base = IntervalTab
        # set bases and create Tab
        Tab.__bases__ = (base,)
        return super(_MetaTab, cls).__call__(*args, **kwargs)


# -- Tab ----------------------------------------------------------------------
# this is the first actual Tab object, all of the functionality is defined
# in the `BaseTab` object

class Tab(BaseTab, metaclass=_MetaTab):
    """A Simple HTML tab.

    This `class` provides a mechanism to generate a full-formatted
    HTML page including banner, navigation-bar, content, and a footer,
    without the user worrying too much about the details.

    For example::

    >>> # import Tab and make a new one with a given title and HTML file
    >>> from gwsumm.tabs import Tab
    >>> tab = Tab('My new tab', 'mytab.html')
    >>> # write the Tab to disk with some simple content
    >>> tab.write_html('This is my content', brand='Brand name')

    Parameters
    ----------
    name : `str`
        name of this tab (required)

    index : `str`
        HTML file in which to write. By default each tab is written to
        an index.html file in its own directory. Use `~Tab.index`
        to find out the default index, if not given.

    shortname : `str`
        shorter name for this tab to use in the navigation bar. By
        default the regular name is used

    parent : `~gwsumm.tabs.Tab`
        parent of this tab. This is used to position this tab in the
        navigation bar.

    children : `list`
        list of child `Tabs <~gwsumm.tabs.Tab>` of this one. This
        is used to position this tab in the navigation bar.

    group : `str`
        name of containing group for this tab in the navigation bar
        dropdown menu. This is only relevant if this tab has a parent.

    path : `str`
        base output directory for this tab (should be the same directory
        for all tabs in this run)

    Notes
    -----
    A `Tab` cannot have both a `~Tab.parent` and `~tab.Children`.
    This is a limitation imposed by the twitter bootstrap navigation bar
    implementation, which does not allow nested dropdown menus. In order
    to collect child tabs in a given place, assign them all the same
    `~Tab.group`.
    """
    type = 'basic'


register_tab(Tab)


class ParentTab(Tab):
    """Dummy `Tab` only for navigation
    """
    def __init__(self, name, children, **kwargs):
        # parse list of children
        if not isinstance(children, list):
            children = [children]
        child = children[0]
        # parse mode and GPS arguments (if required)
        kwargs.setdefault('mode', child.mode)
        if isinstance(self, EventTab):
            kwargs.setdefault('gpstime', child.gpstime)
            kwargs.setdefault('duration', child.duration)
        elif isinstance(self, IntervalTab):
            kwargs.setdefault('span', child.span)
        # create Tab
        super(ParentTab, self).__init__(name, children=children, **kwargs)


# -- TabList -----------------------------------------------------------------

class TabList(list):
    """Custom `list` of `Tab` objects with sorting and parsing
    """

    def __init__(self, entries=[]):
        super(TabList, self).__init__(entries)

    def get_hierarchy(self):
        parents = OrderedDict()
        # 1. Assume all tabs without parents are parents themselves
        for tab in [tab for tab in self if tab.parent is None]:
            parents[tab.name] = tab
        # 2. All remaining tabs without a defined parent define that parent
        # 3. Sort all tabs into their parent sets
        for tab in [tab for tab in self if tab.parent is not None]:
            if tab.parent in parents:
                tab.set_parent(parents[tab.parent])
            elif not isinstance(tab.parent, Tab):
                tab.set_parent(get_tab('default')(
                    tab.parent, mode=tab.mode, span=tab.span))
            parents.setdefault(tab.parent.name, tab.parent)
            if tab not in tab.parent.children:
                tab.parent.add_child(tab)
        return list(parents.values())

    @staticmethod
    def _sortkey(tab):
        # NOTE: we need all return values to be strings for
        #       the sorting to actually work
        if 'Home' in tab.shortname:
            return '1'
        if tab.shortname == 'Summary' and tab.parent is None:
            return '2'
        if tab.shortname == 'Summary':
            return '3'
        if 'ODC' in tab.shortname:
            return '4'
        if tab.shortname.islower():
            return tab.shortname.upper()
        return tab.shortname.lower()

    def sort(self, key=None, reverse=False):
        """Sort this `TabList` in place
        """
        if key is None:
            key = self._sortkey
        hlist = sorted(self.get_hierarchy(), key=key)
        for tab in hlist:
            tab.children.sort(key=key)
        super(TabList, self).sort(key=key, reverse=reverse)

    @classmethod
    def from_ini(cls, config, tag='tab[_-]', match=[],
                 path=os.curdir, plotdir='plots'):
        if isinstance(tag, str):
            tag = re.compile(tag)
        tabs = cls()
        parents = {}
        for section in filter(tag.match, config.sections()):
            # if user gave matches, test match and skip
            if match and section[4:] not in match:
                continue
            # otherwise, get type and create instance of class
            try:
                type_ = config.get(section, 'type')
            except NoOptionError:
                type_ = 'default'
            Tab = get_tab(type_)
            if issubclass(Tab, get_tab('data')):
                tab = Tab.from_ini(config, section, plotdir=plotdir, path=path)
            else:
                tab = Tab.from_ini(config, section, path=path)
            tabs.append(tab)
            if tab.parent and tab.parent.name in parents:
                tab.set_parent(parents[tab.parent.name])
            elif tab.parent:
                parents[tab.parent.name] = tab.parent
        tabs.get_hierarchy()  # call this to resolve map parent names to tabs
        return tabs



# coding=utf-8
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""Definition of the `AccountingTab`
"""

import re
from collections import OrderedDict
from configparser import NoOptionError

from glue.lal import Cache

from gwpy.segments import (DataQualityDict, SegmentList)

from gwdetchar.io import html
from gwdetchar.plot import texify

from ..config import GWSummConfigParser
from .registry import (get_tab, register_tab)
from .. import globalv
from ..data import get_timeseries
from ..segments import get_segments
from ..plot.registry import get_plot
from ..utils import vprint

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['AccountingTab']

ParentTab = get_tab('default')


class AccountingTab(ParentTab):
    """Summarise the data recorded by the operating mode channels
    """
    type = 'accounting'

    @classmethod
    def from_ini(cls, config, section, plotdir='plots', **kwargs):
        new = super(ParentTab, cls).from_ini(config, section, **kwargs)

        # add information
        new.plots = []
        new.channel = config.get(section, 'channel')
        new.ifo = config.get(section, 'IFO')
        new.modes = OrderedDict((int(idx), name) for (idx, name) in
                                config.nditems(section) if idx.isdigit())

        # -----------
        # build plots

        new.set_layout([2, 2])
        new.segmenttag = '%s:%%s' % (new.channel)
        tag = new.channel.split(':', 1)[-1].replace('-', '_')

        groups = type(new.modes)(
            (idx, name.strip('*')) for (idx, name) in new.modes.items()
            if int(idx) % 10 == 0)
        pstates = OrderedDict(
            (idx, name) for (idx, name) in new.modes.items() if
            not name.startswith('*'))

        # parse colors
        colors = dict(
            (int(key[6:]), eval(color)) for (key, color) in
            config.nditems(section) if re.match(r'color-\d+', key))
        for flag in pstates:
            group = int(flag // 10 * 10)
            if flag not in colors:
                colors[flag] = colors.get(group, None)

        # plot segments
        for flags, ptag in zip([groups, pstates], ['overview', 'details']):
            segcolors = [colors.get(flag, None) for flag in flags]
            if not any(segcolors):
                segcolors = None
            for state in new.states:
                new.plots.append(get_plot('segments')(
                    [new.segmenttag % idx for idx in flags],
                    new.span[0], new.span[1], labels=list(flags.values()),
                    state=state, outdir=plotdir,
                    pid='%s_SEGMENTS_%s' % (tag, ptag.upper()),
                    active=segcolors,
                    known={'alpha': 0.1, 'facecolor': 'lightgray'},
                    title='%s operating mode %s' % (new.ifo, ptag)))

        # plot pie charts
        try:
            explode = list(
                map(float, config.get(section, 'pie-explode').split(',')))
        except NoOptionError:
            explode = None
        ptag = 'overview'
        piecolors = [colors.get(flag, None) for flag in groups]
        if not any(piecolors):
            piecolors = None

        for state in new.states:
            labels = list(groups.values())
            new.plots.append(get_plot('segment-pie')(
                [new.segmenttag % idx for idx in groups],
                new.span[0], new.span[1], state=state, labels=labels,
                outdir=plotdir, pid='%s_PIE_%s' % (tag, ptag.upper()),
                colors=piecolors, explode=explode,
                title='%s operating mode %s' % (new.ifo, ptag)))

            shortlabels = ['\n'.join('{0:.9}.'.format(w) if len(w) > 9 else w
                                     for w in l.split()) for
                           l in labels]
            new.plots.append(get_plot('duty')(
                [new.segmenttag % idx for idx in groups],
                new.span[0], new.span[1], state=state, labels=shortlabels,
                outdir=plotdir, pid='%s_SEGMENT_BAR_%s' % (tag, ptag.upper()),
                colors=piecolors, stacked=True, ylim=[0, 100],
                ylabel=texify('Percentage [%] of available time'),
                legend_loc='upper left', legend_bbox_to_anchor=(1., 1),
                legend_fontsize=12, legend_borderaxespad=0,
                legend_frameon=False,
                legend_handlelength=1.0, legend_handletextpad=.5,
                title='%s operating mode %s' % (new.ifo, ptag)))

        return new

    def process(self, nds=None, nproc=1,
                config=GWSummConfigParser(), datacache=None,
                datafind_error='raise', **kwargs):
        """Process time accounting data
        """
        for p in self.plots:
            if p.outputfile in globalv.WRITTEN_PLOTS:
                p.new = False

        # get archived GPS time
        tag = self.segmenttag % list(self.modes)[0]
        try:
            lastgps = globalv.SEGMENTS[tag].known[-1][-1]
        except (IndexError, KeyError):
            lastgps = self.span[0]

        # get data
        new = SegmentList([type(self.span)(lastgps, self.span[1])])
        data = get_timeseries(self.channel, new,
                              config=config, nds=nds, dtype='int16',
                              datafind_error=datafind_error,
                              nproc=nproc, cache=datacache)
        vprint("    All time-series data loaded\n")

        # find segments
        for ts in data:
            modesegments = DataQualityDict()
            for idx, name in self.modes.items():
                # get segments for state
                tag = self.segmenttag % idx
                instate = ts == idx * ts.unit
                modesegments[tag] = instate.to_dqflag(name=name.strip('*'))
                # append segments for group
                group = int(idx // 10. * 10)
                gtag = self.segmenttag % group
                try:
                    modesegments[gtag] += modesegments[tag]
                except KeyError:
                    modesegments[gtag] = modesegments[tag]
            globalv.SEGMENTS += modesegments

        kwargs['segdb_error'] = 'ignore'
        super(AccountingTab, self).process(
            config=config, nds=nds, nproc=nproc,
            segmentcache=Cache(), datacache=datacache,
            datafind_error=datafind_error, **kwargs)

    def write_state_html(self, state):
        """Write the HTML for the given state of this `GuardianTab`
        """
        page = self.scaffold_plots(state=state)

        page.div(class_='row')
        page.div(class_='col-md-12')

        # get segment data
        groups = type(self.modes)((idx, name) for idx, name in
                                  self.modes.items() if idx % 10 == 0)
        modes = type(self.modes)(
            (idx, name) for idx, name in self.modes.items() if
            not name.startswith('*'))
        headers = ['Index', 'Name', 'Active seconds',
                   'Hours', '%']
        prefix = self.ifo.lower()
        for flags, title, id in zip(
                [groups, modes],
                ['Top-level mode information', 'Detailed mode information'],
                ['%s-top-level-mode' % prefix, '%s-detailed-mode' % prefix]):
            page.h1(title, class_='mt-4')
            data = []
            pc = float(abs(state.active) / 100.)
            tots = 0
            toth = 0
            totp = 0
            for idx, name in flags.items():
                flag = get_segments(self.segmenttag % idx,
                                    state.active, query=False).copy()
                actives = abs(flag.active)
                activeh = actives / 3600.
                try:
                    activep = actives / pc
                except ZeroDivisionError:
                    activep = 0
                tots += actives
                toth += activeh
                totp += activep
                data.append([str(idx), name.strip('*'), '%.1f' % actives,
                             '%.1f' % activeh, '%.1f' % activep])
            data.append(['<strong>%s</strong>' % x for x in [
                '', 'Total:', '%.1f' % tots, '%.1f' % toth, '%.1f' % totp]])
            page.add(str(html.table(
                headers, data, id=id,
                caption="%s observatory mode statistics as recorded in "
                        "<samp>%s</samp>" % (title.split()[0], self.channel))))
        return super(ParentTab, self).write_state_html(state, plots=False,
                                                       pre=page)


register_tab(AccountingTab)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for data handling and display
"""

import operator
import warnings
from functools import reduce
from itertools import zip_longest
from collections import OrderedDict

import numpy

from astropy import units

from gwpy.segments import (DataQualityFlag, SegmentList, Segment)
from gwpy.frequencyseries import FrequencySeries
from gwpy.spectrogram import SpectrogramList

from .. import globalv
from ..utils import (vprint, safe_eval)
from ..channels import get_channel
from .utils import (use_segmentlist, get_fftparams, make_globalv_key)
from .timeseries import (get_timeseries, get_timeseries_dict)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@use_segmentlist
def get_coherence_spectrogram(channel_pair, segments, config=None,
                              cache=None, query=True, nds=None,
                              return_=True, frametype=None, nproc=1,
                              datafind_error='raise', return_components=False,
                              **fftparams):
    """Retrieve the time-series and generate a coherence spectrogram of
    the two given channels
    """
    specs = _get_coherence_spectrogram(channel_pair, segments,
                                       config=config, cache=cache,
                                       query=query, nds=nds,
                                       return_=return_, frametype=frametype,
                                       nproc=nproc,
                                       datafind_error=datafind_error,
                                       return_components=return_components,
                                       **fftparams)

    return specs


@use_segmentlist
def _get_coherence_spectrogram(channel_pair, segments, config=None,
                               cache=None, query=True, nds=None,
                               return_=True, frametype=None, nproc=1,
                               datafind_error='raise', return_components=False,
                               **fftparams):

    channel1 = get_channel(channel_pair[0])
    channel2 = get_channel(channel_pair[1])

    # clean fftparams dict using channel 1 default values
    fftparams.setdefault('method', 'welch')
    fftparams = get_fftparams(channel1, **fftparams)

    # key used to store the coherence spectrogram in globalv
    key = make_globalv_key(channel_pair, fftparams)

    # keys used to store component spectrograms in globalv
    components = ('Cxy', 'Cxx', 'Cyy')
    ckeys = [
        make_globalv_key([channel1, channel2], fftparams),
        make_globalv_key(channel1, fftparams),
        make_globalv_key(channel2, fftparams),
    ]

    # convert fftparams to regular dict
    fftparams = fftparams.dict()

    # work out what new segments are needed
    # need to truncate to segments of integer numbers of strides
    stride = float(fftparams.pop('stride'))
    overlap = float(fftparams['overlap'])
    new = type(segments)()
    for seg in segments - globalv.SPECTROGRAMS.get(
            key, SpectrogramList()).segments:
        dur = float(abs(seg)) // stride * stride
        if dur < stride + overlap:
            continue
        new.append(type(seg)(seg[0], seg[0]+dur))

    # extract FFT params for TimeSeries.spectrogram
    spec_fftparams = fftparams.copy()
    for fftkey in ('method', 'scheme',):
        fftparams.pop(fftkey, None)

    # if there are no existing spectrogram, initialize as a list
    globalv.SPECTROGRAMS.setdefault(key, SpectrogramList())

    # XXX HACK: use dummy timeseries to find lower sampling rate
    if len(segments) > 0:
        s = segments[0].start
        dts1 = get_timeseries(channel1, SegmentList([Segment(s, s+1)]),
                              config=config, cache=cache, frametype=frametype,
                              nproc=nproc, query=query,
                              datafind_error=datafind_error, nds=nds)
        dts2 = get_timeseries(channel2, SegmentList([Segment(s, s+1)]),
                              config=config, cache=cache, frametype=frametype,
                              nproc=nproc, query=query,
                              datafind_error=datafind_error, nds=nds)
        try:
            sampling = min(dts1[0].sample_rate.value,
                           dts2[0].sample_rate.value)
        except IndexError:
            sampling = None
    else:
        sampling = None

    # initialize component lists if they don't exist yet
    for ck in ckeys:
        globalv.COHERENCE_COMPONENTS.setdefault(ck, SpectrogramList())

    # get data if query=True or if there are new segments
    query &= abs(new) != 0

    if query:

        # the intersecting segments will be calculated when needed
        intersection = None

        # loop over components needed to calculate coherence
        for comp in components:

            # key used to store this component in globalv (incl sample rate)
            ckey = ckeys[components.index(comp)]

            try:
                filter_ = channel1.frequency_response
            except AttributeError:
                filter_ = None
            else:
                if isinstance(filter_, str):
                    filter_ = safe_eval(filter_, strict=True)

            # check how much of this component still needs to be calculated
            req = new - globalv.COHERENCE_COMPONENTS.get(
                            ckey, SpectrogramList()).segments

            # get data if there are new segments
            if abs(req) != 0:

                # calculate intersection of segments lazily
                # this should occur on first pass (Cxy)
                if intersection is None:
                    total1 = get_timeseries(
                                 channel1, req, config=config,
                                 cache=cache, frametype=frametype,
                                 nproc=nproc, query=query,
                                 datafind_error=datafind_error,
                                 nds=nds)
                    total2 = get_timeseries(
                                 channel2, req, config=config,
                                 cache=cache, frametype=frametype,
                                 nproc=nproc, query=query,
                                 datafind_error=datafind_error,
                                 nds=nds)
                    intersection = total1.segments & total2.segments

                # get required timeseries data (using intersection)
                tslist1, tslist2 = [], []
                if comp in ('Cxy', 'Cxx'):
                    tslist1 = get_timeseries(
                                  channel1, intersection, config=config,
                                  cache=cache, frametype=frametype,
                                  nproc=nproc, query=query,
                                  datafind_error=datafind_error,
                                  nds=nds)
                if comp in ('Cxy', 'Cyy'):
                    tslist2 = get_timeseries(
                                  channel2, intersection, config=config,
                                  cache=cache, frametype=frametype,
                                  nproc=nproc, query=query,
                                  datafind_error=datafind_error,
                                  nds=nds)

                # calculate component
                if len(tslist1) + len(tslist2):
                    vprint("    Calculating component %s for coherence "
                           "spectrogram for %s and %s @ %d Hz" % (
                               comp, str(channel1), str(channel2), sampling))

                for ts1, ts2 in zip_longest(tslist1, tslist2):

                    # ensure there is enough data to do something with
                    if comp in ('Cxx', 'Cxy') and abs(ts1.span) < stride:
                        continue
                    elif comp in ('Cyy', 'Cxy') and abs(ts2.span) < stride:
                        continue

                    # downsample if necessary
                    if ts1 is not None and ts1.sample_rate.value != sampling:
                        ts1 = ts1.resample(sampling)
                    if ts2 is not None and ts2.sample_rate.value != sampling:
                        ts2 = ts2.resample(sampling)

                    # ignore units when calculating coherence
                    if ts1 is not None:
                        ts1._unit = units.Unit('count')
                    if ts2 is not None:
                        ts2._unit = units.Unit('count')

                    # calculate the component spectrogram
                    if comp == 'Cxy':
                        specgram = ts1.csd_spectrogram(
                            ts2, stride, nproc=nproc, **fftparams)
                    elif comp == 'Cxx':
                        specgram = ts1.spectrogram(stride, nproc=nproc,
                                                   **spec_fftparams)
                    elif comp == 'Cyy':
                        specgram = ts2.spectrogram(stride, nproc=nproc,
                                                   **spec_fftparams)

                    if filter_:
                        specgram = (specgram**(1/2.)).filter(*filter_,
                                                             inplace=True) ** 2
                    add_coherence_component_spectrogram(specgram, key=ckey)

                    vprint('.')

                if len(tslist1) + len(tslist2):
                    vprint('\n')

        spans = [SegmentList([  # record spectrogaram spans
            spec.span for spec in globalv.COHERENCE_COMPONENTS[ck]
        ]) for ck in ckeys]
        new = reduce(operator.and_, spans, new).coalesce()
        for seg in new:  # compute coherence from components
            cxy, cxx, cyy = [_get_from_list(
                globalv.COHERENCE_COMPONENTS[ck], seg) for ck in ckeys]
            csg = abs(cxy)**2 / cxx / cyy
            globalv.SPECTROGRAMS[key].append(csg)
            globalv.SPECTROGRAMS[key].coalesce()

    if not return_:
        return

    elif return_components:

        # return list of component spectrogram lists
        out = [SpectrogramList(), SpectrogramList(), SpectrogramList()]
        for comp in components:
            index = components.index(comp)
            ckey = ckeys[index]
            for specgram in globalv.COHERENCE_COMPONENTS[ckey]:
                for seg in segments:
                    if abs(seg) < specgram.dt.value:
                        continue
                    if specgram.span.intersects(seg):
                        common = specgram.span & type(seg)(
                                     seg[0], seg[1] + specgram.dt.value)
                        s = specgram.crop(*common)
                        if s.shape[0]:
                            out[index].append(s)
            out[index] = out[index].coalesce()
        return out

    else:

        # return list of coherence spectrograms
        out = SpectrogramList()
        for specgram in globalv.SPECTROGRAMS[key]:
            for seg in segments:
                if abs(seg) < specgram.dt.value:
                    continue
                if specgram.span.intersects(seg):
                    common = specgram.span & type(seg)(
                                 seg[0], seg[1] + specgram.dt.value)
                    s = specgram.crop(*common)
                    if s.shape[0]:
                        out.append(s)
        return out.coalesce()


def get_coherence_spectrum(channel_pair, segments, config=None,
                           cache=None, query=True, nds=None, return_=True,
                           **fftparams):
    """Retrieve the time-series and generate a coherence spectrogram of the
    given channel
    """

    channel1 = get_channel(channel_pair[0])
    channel2 = get_channel(channel_pair[1])

    if isinstance(segments, DataQualityFlag):
        name = ','.join([channel1.ndsname, channel2.ndsname, segments.name])
        segments = segments.active
    else:
        name = channel1.ndsname + ',' + channel2.ndsname
    name += ',%s' % format
    cmin = name + '.min'
    cmax = name + '.max'

    if name not in globalv.COHERENCE_SPECTRUM:
        vprint("    Calculating 5/50/95 percentile spectra for %s"
               % name.rsplit(',', 1)[0])

        # ask for a list of component spectrograms (a list of SpectrogramLists)
        speclist = get_coherence_spectrogram(
                       channel_pair, segments, config=config, cache=cache,
                       query=query, nds=nds, return_components=True,
                       **fftparams)

        cdict = {}
        components = ('Cxy', 'Cxx', 'Cyy')

        # join spectrograms in each list so we can average it
        for comp in components:
            index = components.index(comp)
            try:
                cdict[comp] = speclist[index].join(gap='ignore')
            except ValueError as e:
                if 'units do not match' in str(e):
                    warnings.warn(str(e))
                    for spec in speclist[index][1:]:
                        spec.unit = speclist[0].unit
                    cdict[comp] = speclist[index].join(gap='ignore')
                else:
                    raise

        # average spectrograms to get PSDs and CSD
        try:
            Cxy = complex_percentile(cdict['Cxy'], 50)
            Cxx = cdict['Cxx'].percentile(50)
            Cyy = cdict['Cyy'].percentile(50)
            globalv.COHERENCE_SPECTRUM[name] = FrequencySeries(
                numpy.abs(Cxy)**2 / Cxx / Cyy, f0=Cxx.f0, df=Cxx.df)
        except (ValueError, IndexError):
            globalv.COHERENCE_SPECTRUM[name] = FrequencySeries(
                [], channel=channel1, f0=0, df=1, unit=units.Unit(''))
            globalv.COHERENCE_SPECTRUM[cmin] = globalv.COHERENCE_SPECTRUM[name]
            globalv.COHERENCE_SPECTRUM[cmax] = globalv.COHERENCE_SPECTRUM[name]
        else:
            # FIXME: how to calculate percentiles correctly?
            globalv.COHERENCE_SPECTRUM[cmin] = FrequencySeries(
                abs(complex_percentile(cdict['Cxy'], 5))**2 /
                cdict['Cxx'].percentile(95) / cdict['Cyy'].percentile(95),
                f0=Cxx.f0, df=Cxx.df)
            globalv.COHERENCE_SPECTRUM[cmax] = FrequencySeries(
                abs(complex_percentile(cdict['Cxy'], 95))**2 /
                cdict['Cxx'].percentile(5) / cdict['Cyy'].percentile(5),
                f0=Cxx.f0, df=Cxx.df)

        # set the spectrum's name manually; this will be used for the legend
        globalv.COHERENCE_SPECTRUM[name].name = (
            channel1.ndsname + '\n' + channel2.ndsname)

        vprint(".\n")

    if not return_:
        return

    cmin = '%s.min' % name
    cmax = '%s.max' % name
    out = (globalv.COHERENCE_SPECTRUM[name], globalv.COHERENCE_SPECTRUM[name],
           globalv.COHERENCE_SPECTRUM[name])
    return out


def add_coherence_component_spectrogram(specgram, key=None, coalesce=True):
    """Add a `Coherence spectrogram` to the global memory cache
    """
    if key is None:
        key = specgram.name or str(specgram.channel)
    globalv.COHERENCE_COMPONENTS.setdefault(key, SpectrogramList())
    globalv.COHERENCE_COMPONENTS[key].append(specgram)
    if coalesce:
        globalv.COHERENCE_COMPONENTS[key].coalesce()


@use_segmentlist
def get_coherence_spectrograms(channel_pairs, segments, config=None,
                               cache=None, query=True, nds=None,
                               return_=True, frametype=None, nproc=1,
                               datafind_error='raise', **fftparams):
    """Get coherence spectrograms for multiple channels
    """
    if fftparams.get('method', 'welch') != 'welch':
        raise ValueError("Cannot process coherence data with method=%r"
                         % fftparams.get('method'))
    fftparams['method'] = 'welch'
    channels = list(map(get_channel, channel_pairs))
    pairs = list(zip(channels[0::2], channels[1::2]))

    # get timeseries data in bulk
    if query:
        qchannels = []
        havesegs = []
        for c1, c2 in pairs:
            c1 = get_channel(c1)
            c2 = get_channel(c2)
            fftparams_ = get_fftparams(c1, **fftparams)
            key = make_globalv_key((c1, c2), fftparams_)
            qchannels.extend((c1, c2))
            havesegs.append(globalv.SPECTROGRAMS.get(
                key, SpectrogramList()).segments)
        havesegs = reduce(operator.and_, havesegs)
        new = segments - havesegs
        strides = set([getattr(c, 'stride', 0) for c in qchannels])
        if len(strides) == 1:
            stride = strides.pop()
            new = type(new)([s for s in new if abs(s) >= stride])
        get_timeseries_dict(qchannels, new, config=config, cache=cache,
                            nproc=nproc, frametype=frametype,
                            datafind_error=datafind_error, nds=nds,
                            return_=False)
    # loop over channels and generate spectrograms
    out = OrderedDict()

    for channel_pair in pairs:
        out[channel_pair] = get_coherence_spectrogram(
            channel_pair, segments, config=config, cache=cache, query=query,
            nds=nds, nproc=nproc, return_=return_,
            datafind_error=datafind_error, **fftparams)
    return out


def _get_from_list(serieslist, segment):
    """Internal function to crop a series from a serieslist

    Should only be used in situations where the existence of the target
    data within the list is guaranteed
    """
    for series in serieslist:
        if segment.intersects(series.span):
            outseg = segment & series.span
            return series.crop(*outseg)
    raise ValueError("Cannot crop series for segment %s from list"
                     % str(segment))


def complex_percentile(array, percentile):
    re = array.real.percentile(percentile)
    im = array.imag.percentile(percentile) * 1j
    return re + im



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Thin wrapper around configparser
"""

import configparser
import os
import re
import sys

# import these for evaluating lambda expressions in the configuration file
import math  # noqa: F401
import numpy  # noqa: F401

from collections import OrderedDict
from importlib import import_module
from io import StringIO
from http.client import HTTPException

from matplotlib import rcParams

from astropy import units

from gwpy.detector import (Channel, ChannelList)
from gwpy.time import tconvert

from .channels import (get_channels, split as split_channels,
                       update_channel_params)
from .html import (get_css, get_js)
from .utils import (nat_sorted, re_cchar, re_quote, safe_eval, OBSERVATORY_MAP)

__all__ = [
    'GWSummConfigParser',
]

CONFIGDIR = os.path.join(sys.prefix, 'etc', 'gwsumm', 'configuration')


# -- class definitions --------------------------------------------------------

class GWSummConfigParser(configparser.ConfigParser):
    # preserve case in options
    optionxform = str
    # disable colon separator
    OPTCRE = re.compile(
        r'(?P<option>[^=\s][^=]*)\s*(?P<vi>[=])\s*(?P<value>.*)$')
    # set interpolation match
    _interpvar_re = re.compile(r"%\(([^)]+)\)s")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('dict_type', OrderedDict)
        configparser.ConfigParser.__init__(self, *args, **kwargs)
    __init__.__doc__ = configparser.ConfigParser.__init__.__doc__

    def ndoptions(self, section, **kwargs):
        options = configparser.ConfigParser.options(self, section, **kwargs)
        return [o for o in options if o not in self._defaults]

    def nditems(self, section, **kwargs):
        items = configparser.ConfigParser.items(self, section, **kwargs)
        return [i for i in items if i[0] not in self._defaults]

    def read(self, filenames):
        readok = configparser.ConfigParser.read(self, filenames)
        if isinstance(filenames, str):
            filenames = filenames.split(',')
        for fp in filenames:
            if fp not in readok:
                raise IOError("Cannot read file: %s" % fp)
        self.files = list(map(os.path.abspath, readok))
        return readok

    @classmethod
    def from_configparser(cls, cp):
        """Copy an existing :class:`~configparser.ConfigParser`.
        """
        # if its already what we need, just return this instance
        if isinstance(cp, cls):
            return cp
        # set up temporary buffer
        buf = StringIO()
        # write to buffer
        cp.write(buf)
        buf.seek(0)
        # read new GWSummConfigParser
        new = cls()
        new.read_file(buf)
        return new

    def __repr__(self):
        return '<GWSummConfigParser()>'

    def interpolate_section_names(self, **kwargs):
        """Interpolate a specific key in a section name using val
        """
        for section in self.sections():
            s = section
            for key in self._interpvar_re.findall(section):
                try:
                    val = kwargs[key]
                except KeyError:
                    raise configparser.InterpolationMissingOptionError(
                        '[%s]' % section, section, key, '%%(%s)s' % key)
                s = section % {key: val}
            self._sections[s] = self._sections.pop(section)

    def set_ifo_options(self, ifo, observatory=None,
                        section=configparser.DEFAULTSECT):
        """Set configurations options in [DEFAULT] based on the given `ifo`

        The following options are set

        - `IFO` - the two-character interferometer prefix, e.g. ``L1``
        - `ifo` - the two-character interferometer prefix in lower-case,
           e.g. ``l1``
        - `SITE` - the single-character site ID, e.g. ``L``
        - `site` - the single-character site ID,n lower-case e.g. ``l``

        Additionally, if `observatory` is given, or the `ifo` matches known
        observatories, the following option is set

        - `observatory` - the name of the observatory, e.g. ``LIGO Livingston``

        """
        if observatory is None:
            observatory = OBSERVATORY_MAP.get(ifo)
        self.set(section, 'IFO', ifo)
        self.set(section, 'ifo', ifo.lower())
        self.set(section, 'SITE', ifo[0].upper())
        self.set(section, 'site', ifo[0].lower())
        if observatory is not None:
            self.set(section, 'observatory', observatory)

    def set_date_options(self, start, end, section=configparser.DEFAULTSECT):
        """Set datetime options in [DEFAULT] based on the given times

        The following options are set

        - `gps-start-time` - the integer GPS start time of this job
        - `gps-end-time` - the integer GPS end time of this job
        - `yyyy` - the four-digit year of the start date
        - `mm` - the two-digit month of the start date
        - `dd` - the two-digit day-of-month of the start date
        - `yyyymm` - the six-digit year and month of the start date
        - `yyyymmdd` - the eight-digit year-month-day of the start date
        - `duration` - the duration of the job (seconds)

        Additionally, if LAL is available, the following extra options
        are also set

        - `leap-seconds` - the number of leap seconds for the start date
        - `gps-start-time-noleap` - the leap-corrected integer GPS start time
        - `gps-end-time-noleap` - the leap-corrected integer GPS end time

        """
        utc = tconvert(start)
        self.set(section, 'gps-start-time', str(int(start)))
        self.set(section, 'gps-end-time', str(int(end)))
        self.set(section, 'yyyy', utc.strftime('%Y'))
        self.set(section, 'yy', utc.strftime('%y'))
        self.set(section, 'mm', utc.strftime('%m'))
        self.set(section, 'dd', utc.strftime('%d'))
        self.set(section, 'yyyymm', utc.strftime('%Y%m'))
        self.set(section, 'yyyymmdd', utc.strftime('%Y%m%d'))
        self.set(section, 'duration', str(int(end - start)))
        try:
            from lal import GPSLeapSeconds as leap_seconds
        except ImportError:
            pass
        else:
            nleap = leap_seconds(int(start))
            self.set(section, 'leap-seconds', str(nleap))
            self.set(section, 'gps-start-time-noleap', str(int(start) - nleap))
            self.set(section, 'gps-end-time-noleap', str(int(end) - nleap))

    def finalize(self):
        """Finalize this `GWSummConfigParser` by running all the loaders

        This method is just a shortcut to run each of the following

        .. autosummary::

           ~GWSummConfigParser.load_plugins
           ~GWSummConfigParser.load_units
           ~GWSummConfigParser.load_channels
           ~GWSummConfigParser.load_states

        """
        self.load_plugins()
        self.load_units()
        self.load_channels()
        self.load_states()

    def load_plugins(self):
        """Load all plugin modules as defined in the [plugins] section
        """
        mods = []
        try:
            plugins = self.ndoptions('plugins')
        except configparser.NoSectionError:
            pass
        else:
            for plugin in plugins:
                mods.append(import_module(plugin))
        return mods

    def load_units(self):
        """Load all unit definitions as defined in the [units] section
        """
        try:
            customunits = self.nditems('units')
        except configparser.NoSectionError:
            return []
        else:
            new_ = []
            for unit, b in customunits:
                if b.lower() == 'dimensionless':
                    b = ''
                new_.append(units.def_unit([unit], units.Unit(b)))
            units.add_enabled_units(new_)
            return new_

    def load_channels(self):
        """Load all channel definitions as given in the selfuration

        Channels are loaded from sections named [channels-...] or
        those sections whose name is a channel name in itself
        """
        channels_re = re.compile(r'channels[-\s]')
        # parse channel grouns into individual sections
        for section in filter(channels_re.match, self.sections()):
            names = split_channels(self.get(section, 'channels'))
            items = dict(self.nditems(section, raw=True))
            items.pop('channels')
            for name in names:
                name = name.strip(' \n')
                if not self.has_section(name):
                    self.add_section(name)
                for key, val in items.items():
                    if not self.has_option(name, key):
                        self.set(name, key, val)

        # read all channels
        raw = set()
        trend = set()
        for section in self.sections():
            try:
                m = Channel.MATCH.match(section).groupdict()
            except AttributeError:
                continue
            else:
                if not m['ifo']:
                    continue
            if m['trend']:
                trend.add(section)
            else:
                raw.add(section)

        channels = ChannelList()
        for group in [raw, trend]:
            try:
                newchannels = get_channels(group)
            except HTTPException:
                newchannels = []

            # read custom channel definitions
            for channel, section in zip(newchannels, group):
                for key, val in nat_sorted(self.nditems(section),
                                           key=lambda x: x[0]):
                    key = re_cchar.sub('_', key.rstrip())
                    if key.isdigit():
                        if not hasattr(channel, 'bits'):
                            channel.bits = []
                        while len(channel.bits) < int(key):
                            channel.bits.append(None)
                        if val.startswith('r"') or val.startswith('r\''):
                            val = eval(val)
                        channel.bits.append(str(val))
                    else:
                        setattr(channel, key, safe_eval(val.rstrip()))
                channels.append(channel)

        # rebuild channel list with new parameters
        update_channel_params()

        return channels

    def load_states(self, section='states'):
        """Read and format a list of `SummaryState` definitions from the
        given :class:`~configparser.ConfigParser`
        """
        from .state import (register_state, SummaryState,
                            ALLSTATE, generate_all_state, get_state)
        # parse the [states] section into individual state definitions
        try:
            states = dict(self.nditems(section))
        except configparser.NoSectionError:
            self.add_section(section)
            states = {}
        for state in states:
            if not (self.has_section('state-%s' % state) or
                    self.has_section('state %s' % state)):
                section = 'state-%s' % state
                self.add_section(section)
                self.set(section, 'name', state)
                self.set(section, 'definition', states[state])

        # parse each state section into a new state
        states = []
        for section in self.sections():
            if re.match(r'state[-\s]', section):
                states.append(register_state(
                    SummaryState.from_ini(self, section)))

        # register All state
        start = self.getint(section, 'gps-start-time')
        end = self.getint(section, 'gps-end-time')
        try:
            all_ = generate_all_state(start, end)
        except ValueError:  # user defined an all state themselves
            all_ = get_state(ALLSTATE)
        else:
            states.insert(0, all_)

        return states

    def load_rcParams(self, section='rcParams'):
        """Load custom `matplotlib.rcParams` for plots in this analysis
        """
        try:
            new = dict(self.nditems(section))
        except configparser.NoSectionError:
            return dict()
        else:
            for key, value in new.items():
                new[key] = safe_eval(value)
            rcParams.update(new)
            return new

    def get_css(self, section='html'):
        # get critical CSS
        css = get_css().copy()
        for key in css:
            try:
                css[key] = self.get(section, '%s-css' % key)
            except configparser.NoSectionError:  # no overrides are present
                return list(css.values())
            except configparser.NoOptionError:
                continue
        files = list(css.values())
        # get extra CSS
        try:
            extras = self.get(section, 'extra-css')
        except configparser.NoOptionError:
            return files
        else:
            files.extend([re_quote.sub('', x) for x in extras.split(',')])
        return files

    def get_javascript(self, section='html'):
        # get critical JS
        js = get_js().copy()
        for key in js:
            try:
                js[key] = self.get(section, '%s-js' % key)
            except configparser.NoSectionError:
                return list(js.values())
            except configparser.NoOptionError:
                continue
        files = list(js.values())
        # get extra JS
        try:
            extras = self.get(section, 'extra-js')
        except configparser.NoOptionError:
            return files
        else:
            files.extend([re_quote.sub('', x) for x in extras.split(',')])
        return files



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of the GWSumm package.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the GWSumm package
"""

import glob
import os.path
import versioneer

from setuptools import setup

version = versioneer.get_version()
cmdclass = versioneer.get_cmdclass()

try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    pass
else:
    cmdclass["build_sphinx"] = BuildDoc

# configuration files
data_files = [
    ("etc/gwsumm/configuration",
     glob.glob(os.path.join("share", "*.ini"))),
]

# run setup
# NOTE: all other metadata and options come from setup.cfg
setup(
    version=version,
    project_urls={
         "Bug Tracker": "https://github.com/gwpy/gwsumm/issues",
         "Discussion Forum": "https://gwdetchar.slack.com",
         "Documentation": "https://gwsumm.readthedocs.io",
         "Source Code": "https://github.com/gwpy/gwsumm",
     },
    cmdclass=cmdclass,
    data_files=data_files,
)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Job modes
"""

import os.path
from enum import (Enum, unique)

from . import globalv

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- operating Mode -----------------------------------------------------------

# https://docs.python.org/3/library/enum.html#orderedenum
class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


# set mode enum
@unique
class Mode(OrderedEnum):
    """Enumeration of valid processing 'modes'

    Each mode provides an association with a particular GPS interval
    """
    # no GPS associations
    static = 0
    # central GPS time (with duration)
    event = 1
    # arbitrary GPS [start, end) interval
    gps = 2
    # calendar epochs
    day = 10
    week = 11
    month = 12
    year = 13

    def dir_format(self):
        if self == Mode.day:
            return os.path.join('day', '%Y%m%d')
        elif self == Mode.week:
            return os.path.join('week', '%Y%m%d')
        elif self == Mode.month:
            return os.path.join('month', '%Y%m')
        elif self == Mode.year:
            return os.path.join('year', '%Y')
        raise ValueError("Cannot format base for Mode %s" % self)

    def is_calendar(self):
        if self >= Mode.day:
            return True
        return False


# -- Mode accessors -----------------------------------------------------------

def get_mode(m=None):
    """Return the enum for the given mode, defaults to the current mode.
    """
    if m is None:
        m = globalv.MODE
    if isinstance(m, (int, Enum)):
        return Mode(m)
    else:
        try:
            return Mode[str(m).lower()]
        except KeyError:
            raise ValueError("%s is not a valid Mode" % m)


def set_mode(m):
    """Set the current mode.
    """
    if isinstance(m, int):
        m = Mode(m)
    elif not isinstance(m, Mode):
        try:
            m = Mode[str(m).lower()]
        except KeyError:
            raise ValueError("%s is not a valid Mode" % m)
    globalv.MODE = m.value


# -- Mode utilities -----------------------------------------------------------

def get_base(date, mode=None):
    """Determine the correct base attribute for the given date and mode.

    Parameters
    ----------
    date : :class:`datetime.datetime`
        formatted date
    mode : `int`, `str`
        enumerated interger code (or name) for the required mode

    Returns
    -------
    base : `str`
        the recommended base URL to have a correctly linked calendar
    """
    mode = get_mode(mode)
    return date.strftime(mode.dir_format())



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwsumm.plot.guardian` command-line interface
"""

import os
import shutil

from gwpy.timeseries import (
    TimeSeries,
    TimeSeriesList,
)

from .... import globalv
from ....archive import write_data_archive
from .. import __main__ as guardian_cli

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- test configuration

CONFIG = """
[tab-ISC_LOCK]
type = guardian
node = ISC_LOCK
name = %(node)s
; node states
600 = Low noise
"""

# -- test data

SUFFICES = [
    "STATE_N",
    "REQUEST_N",
    "NOMINAL_N",
    "OK",
    "MODE",
    "OP",
]
DATA = {
    key: TimeSeriesList(
        TimeSeries([600] * 3600 * 16, sample_rate=16, name=key, channel=key)
    ) for key in ["L1:GRD-ISC_LOCK_{}".format(suff) for suff in SUFFICES]
}


# -- utils --------------------------------------------------------------------

def _get_inputs(workdir):
    """Prepare, and return paths to, input data products
    """
    # set global timeseries data
    globalv.DATA = DATA
    # get path to data files
    ini = os.path.join(workdir, "config.ini")
    archive = os.path.abspath(os.path.join(workdir, "archive.h5"))
    # write to data files
    with open(ini, 'w') as f:
        f.write(CONFIG)
    write_data_archive(archive)
    # reset global data and return
    globalv.DATA = {}
    return (ini, archive)


# -- cli tests ----------------------------------------------------------------

def test_main(tmpdir, caplog):
    outdir = str(tmpdir)
    plot = os.path.join(outdir, "guardian.png")
    (ini, archive) = _get_inputs(outdir)
    args = [
        'ISC_LOCK',
        '0', '3600',
        ini,
        '--plot-params', 'title=Test figure',
        '--output-file', plot,
        '--verbose',
        '--archive', archive,
    ]
    # test output
    guardian_cli.main(args)
    assert os.path.exists(plot)
    assert len(os.listdir(outdir)) == 3  # 2 inputs, 1 output
    assert 'Read data archive from {}'.format(archive) in caplog.text
    assert 'Processing:' in caplog.text
    assert 'Plot saved to {}'.format(plot) in caplog.text
    assert 'Archive recorded as {}'.format(archive) in caplog.text
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for twitter-bootstrap HTML constructs.
"""

from MarkupPy import markup

from gwdetchar.io import html

from ..mode import (Mode, get_mode)
from ..utils import re_cchar

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set <meta> for bootstrap
META = {'viewport': 'width=device-width, initial-scale=1.0'}


# -----------------------------------------------------------------------------
# variable HTML constructs

def banner(title, subtitle=None, titleclass=None, subtitleclass=None):
    """Construct a banner heading in bootstrap format

    Parameters
    ----------
    title : `str`
        name of page (<h1>)
    subtitle : `str`, optional
        description of page (<p>)
    titleclass : `str`, optional
        class option for <h1>
    subtitleclass : `str`, optional
        class option for <p>

    Returns
    -------
    banner : `~MarkupPy.markup.page`
        markup.py `page` instance
    """
    page = markup.page()
    page.div(class_='banner')
    if titleclass is None:
        page.h1(str(title))
    else:
        page.h1(str(title), class_=titleclass)
    if subtitle is not None and subtitleclass is not None:
        page.p(subtitle, class_=subtitleclass)
    elif subtitle is not None:
        page.p(subtitle)
    page.div.close()

    return page


def calendar(date, tag='a', class_='nav-link dropdown-toggle',
             id_='calendar', dateformat=None, mode=None):
    """Construct a bootstrap-datepicker calendar.

    Parameters
    ----------
    date : :class:`datetime.datetime`, :class:`datetime.date`
        active date for the calendar
    tag : `str`
        type of enclosing HTML tag, default: ``<a>``

    Returns
    -------
    calendar : `list`
        a list of three oneliner strings of HTML containing the calendar
        text and a triggering dropdown
    """
    mode = get_mode(mode)
    if dateformat is None:
        if mode == Mode.day:
            dateformat = '%B %d %Y'
        elif mode == Mode.week:
            dateformat = 'Week of %B %d %Y'
        elif mode == Mode.month:
            dateformat = '%B %Y'
        elif mode == Mode.year:
            dateformat = '%Y'
        else:
            raise ValueError("Cannot generate calendar for Mode %s" % mode)
    datestring = date.strftime(dateformat).replace(' 0', ' ')
    data_date = date.strftime('%d-%m-%Y')
    # get navigation objects
    backward = markup.oneliner.a(
        '&laquo;', class_='nav-link step-back', title='Step backward')
    cal = markup.oneliner.a(
        datestring, id_=id_, class_=class_, title='Show/hide calendar',
        **{'data-date': data_date, 'data-date-format': 'dd-mm-yyyy',
           'data-viewmode': '%ss' % mode.name})
    forward = markup.oneliner.a(
        '&raquo;', class_='nav-link step-forward', title='Step forward')
    return [backward, cal, forward]


def wrap_content(page):
    """Utility to wrap some HTML into the relevant <div>s for the main
    body of a page in bootstrap format

    Parameters
    ----------
    page : :class:`~MarkupPy.markup.page`, `str`
        HTML content to be wrapped
    span : `int`
        column span of content, default: 'full' (``12``)

    Returns
    -------
    wrappedpage : :class:`~MarkupPy.markup.page`
        A new `page` with the input content wrapped as

        .. code:: html

            <div class="container">
            </div>
    """
    out = markup.page()
    out.div(class_='container-fluid', id_='main')
    out.add(str(page))
    out.div.close()
    return out


def state_switcher(states, default=0):
    """Build a state switch button, including all of the given
    states, with the default selected by index
    """
    current, chref = states[default]
    page = markup.page()
    page.ul(class_='nav navbar-nav')
    page.li(class_='nav-item dropdown')
    page.a(str(current), class_='nav-link dropdown-toggle', href='#',
           id_='states', role='button', title='Show/hide state menu',
           **{'data-toggle': 'dropdown'})
    page.div(
        class_='dropdown-menu dropdown-menu-right state-switch shadow',
        id_='statemenu',
    )
    page.h6('Select below to view this page in another state (different '
            'time segments).', class_='dropdown-header')
    page.div('', class_='dropdown-divider')
    for i, (state, href) in enumerate(states):
        page.a(str(state), class_='dropdown-item state', title=str(state),
               id_='state_%s' % re_cchar.sub('_', str(state)).lower(),
               onclick='jQuery(this).load_state(\'%s\');' % href)
    page.div.close()  # dropdown-menu dropdown-menu-right
    page.li.close()  # nav-item dropdown state-switch
    page.ul.close()  # nav navbar-nav
    return page


def base_map_dropdown(this, id_=None, bases=dict()):
    """Construct a dropdown menu that links to a version of the current
    page on another server, based on a new base.
    """
    # format id
    if id_:
        id_ = dict(id_=id_)
    else:
        id_ = dict()
    # format links
    baselinks = [markup.oneliner.a(
        key, title=key, class_='dropdown-item', **{'data-new-base': val}
    ) for (key, val) in bases.items() if key != this]
    # slam it all together
    page = markup.page()
    if baselinks:
        page.div(class_='dropdown base-map', **id_)
        page.add(str(html.dropdown(
            this,
            baselinks,
            class_='navbar-brand nav-link border border-white '
                   'rounded dropdown-toggle',
        )))
        page.div.close()
    else:
        page.div(
            str(this),
            class_='navbar-brand border border-white rounded',
            **id_,
        )
    return page



# coding=utf-8
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""Definition of the `GuardianTab`
"""

import re
from collections import OrderedDict
from configparser import NoOptionError

from dateutil import tz

import numpy

from matplotlib.cm import get_cmap
from matplotlib.colors import Normalize

from astropy.time import Time

from glue.lal import Cache

from gwpy.segments import DataQualityDict

from gwdetchar.io import html
from gwdetchar.plot import texify

from .. import globalv
from ..config import GWSummConfigParser
from ..data import get_timeseries_dict
from ..plot.registry import get_plot
from ..segments import get_segments
from ..state import ALLSTATE
from ..utils import vprint
from .registry import (get_tab, register_tab)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['GuardianTab']

DataTab = get_tab('default')
UTC = tz.gettz('UTC')
REQUESTSTUB = '+request'
NOMINALSTUB = '+nominal'
MODE_COLORS = ['grey', 'magenta', 'red', 'saddlebrown']

re_guardian_index = re.compile(r'\[(?P<idx>.*)\] (?P<label>.*)')


class GuardianTab(DataTab):
    """Summarises the data recorded by an Advanced LIGO Guardian node.

    Each guardian node controls and monitors state transitions for a
    specific subsystem of the Advanced LIGO interferometer. The
    `GuardianTab` summarises those data with a state segments plot, a
    transitions summary table, and a detailed list of transitions and
    segments for each listed state.
    """
    type = 'guardian'

    @classmethod
    def from_ini(cls, config, section, plotdir='plots', **kwargs):
        """Define a new `GuardianTab`.
        """
        new = super(DataTab, cls).from_ini(config, section, **kwargs)
        new.plots = []
        new.plotdir = plotdir
        new.ifo = config.get(section, 'IFO')

        # record node and states
        new.node = config.get(section, 'node')
        new.grdstates = OrderedDict()
        plot = []
        for key, name in config.nditems(section):
            try:
                key = int(key)
            except ValueError:
                continue
            else:
                if name[0] == '*':
                    plot.append(False)
                else:
                    plot.append(True)
                new.grdstates[int(key)] = name.strip('*')
        try:
            new.transstates = list(map(
                int, config.get(section, 'transitions').split(',')))
        except NoOptionError:
            new.transstates = list(new.grdstates)

        # -- build plots ------------------------

        new.set_layout([1, 2])
        grdidxs = dict((state, idx) for idx, state in
                       new.grdstates.items())
        new.segmenttag = '%s:%s %%s' % (new.ifo, new.node)
        pstates = [l for i, l in enumerate(list(new.grdstates.values())[::-1])
                   if plot[-i-1]]
        flags = [new.segmenttag % name for name in pstates]
        labels = ['[%d] %s' % (grdidxs[state], state) for state in pstates]

        # define colours
        cmap = get_cmap('brg')(Normalize(-1, 1)(
            numpy.linspace(-1, 1, len(pstates))))[::-1]
        th = len(flags) > 8 and (new.span[1] - new.span[0])/200. or 0

        for state in new.states:
            # get common plot tag prefix
            tagprefix = 'GRD_%s' % re.sub(r'[-\s]', '_', new.node.upper())
            if state.name.lower() != ALLSTATE:  # include state name if not All
                tagprefix = '%s_%s' % (state.name, tagprefix)
            # segment plot
            new.plots.append(get_plot('guardian')(
                flags, new.span[0], new.span[1], state=state,
                labels=labels, outdir=plotdir,
                known={'hatch': 'x', 'alpha': 0.1, 'facecolor': 'none'},
                tag='%s_SEGMENTS' % tagprefix,
                title='%s Guardian %s state' % (
                    new.ifo, texify(new.node)), zorder=2))

            # pie
            new.plots.append(get_plot('segment-pie')(
                flags, new.span[0], new.span[1], state=state,
                labels=pstates, colors=cmap,
                tag='%s_SEGMENT_PIE' % tagprefix,
                startangle=180, counterclock=False, wedge_linewidth=0.01,
                outdir=plotdir, title='%s Guardian %s state' % (
                    new.ifo, texify(new.node)),
                legend_fontsize=16, legend_sorted=True, legend_threshold=th,
            ))

            # bar
            new.plots.append(get_plot('segment-bar')(
                flags, new.span[0], new.span[1], state=state,
                labels=pstates, sorted=True,
                tag='%s_SEGMENT_BAR' % tagprefix,
                outdir=plotdir, title='%s Guardian %s state' % (
                    new.ifo, texify(new.node)),
            ))
        return new

    def process(self, nds=None, nproc=1,
                config=GWSummConfigParser(), datacache=None,
                segmentcache=Cache(), datafind_error='raise', **kwargs):
        """Process data for the given state.
        """
        # finalize state information
        self.finalize_states(
            config=config, segdb_error=kwargs.get('segdb_error', 'raise'),
            datafind_error=datafind_error)
        vprint("States finalised [%d total]\n" % len(self.states))
        vprint("    Default state: %r\n" % str(self.defaultstate))

        # remove plots that have already been generated
        for p in self.plots:
            if p.outputfile in globalv.WRITTEN_PLOTS:
                p.new = False

        # --------------------------------------------------------------------
        # work out which channels are needed

        prefix = '%s:GRD-%s_%%s' % (self.ifo, self.node)
        state = sorted(self.states, key=lambda s: abs(s.active))[-1]

        prefices = ['STATE_N', 'REQUEST_N', 'NOMINAL_N', 'OK', 'MODE', 'OP']
        alldata = list(get_timeseries_dict(
            [prefix % x for x in prefices],
            state, config=config, nds=nds, nproc=nproc,
            cache=datacache, datafind_error=datafind_error,
            dtype='int32').values())
        vprint("    All time-series data loaded\n")

        # --------------------------------------------------------------------
        # find segments and transitions

        self.transitions = dict((v, []) for v in self.grdstates)

        for sdata, rdata, ndata, okdata in zip(*alldata[:4]):
            ssegs = DataQualityDict()
            rsegs = DataQualityDict()
            nsegs = DataQualityDict()
            oksegs = (okdata == 1).to_dqflag(name='Node OK')
            for v, name in self.grdstates.items():
                # get segments for state
                tag = self.segmenttag % name
                instate = sdata == v
                ssegs[tag] = instate.to_dqflag(name=name)
                diff_ = numpy.diff(instate.value.astype(int))
                transin = (diff_ == 1).nonzero()[0] + 1
                transout = (diff_ == -1).nonzero()[0] + 1
                for i, j in zip(transin, transout):
                    t = sdata.times[i].value
                    from_ = sdata[i-1].value
                    to_ = sdata[j].value
                    self.transitions[v].append((t, from_, to_))
                # get segments for request
                tag = self.segmenttag % name + REQUESTSTUB
                instate = rdata == v
                rsegs[tag] = instate.to_dqflag(name=name)
                # get segments for nominal
                tag = self.segmenttag % name + NOMINALSTUB
                nom = ndata == v
                nsegs[tag] = nom.to_dqflag(name=name)

            globalv.SEGMENTS += ssegs
            globalv.SEGMENTS += rsegs
            globalv.SEGMENTS += nsegs
            globalv.SEGMENTS += {self.segmenttag % 'OK': oksegs}

        super(GuardianTab, self).process(
            config=config, nds=nds, nproc=nproc,
            datacache=datacache, segmentcache=segmentcache, **kwargs)

    def write_state_html(self, state):
        """Write the HTML for the given state of this `GuardianTab`
        """
        page = self.scaffold_plots(state=state)

        page.div(class_='row')
        page.div(class_='col-md-12')
        page.h2('%s state transitions' % self.node, class_='mt-4 mb-4')
        page.add(html.alert(  # add dismissable alert
            'For all of the following data, "Unknown" simply labels any '
            'state in this node that was not chosen for display, and does '
            'not mean that the state was unrecognised by the Guardian '
            'system. Transitions from an "Unkown" state are not listed in '
            'the table below, but are included in the totals.',
            context=self.ifo.lower()))

        # draw table
        id_ = '{}-state-transitions'.format(self.ifo.lower())
        page.table(class_='table table-sm table-hover transitions mt-2',
                   id_=id_)
        page.caption('Transitions into each state (row) from each other '
                     'state (column). Only those states named in the '
                     'configuration are shown, but the \'Total\' includes '
                     'transitions from any and all states. A dash '
                     '(\'&mdash;\') indicates no transitions from '
                     'the given state.')
        page.thead()
        page.tr()
        for th in ['State'] + list(self.grdstates.values()) + ['Total']:
            page.th(th)
        page.tr.close()
        page.thead.close()
        page.tbody()
        for i, bit in enumerate(self.transstates):
            page.tr()
            name = self.grdstates[bit].strip('*')
            page.th(name)
            for j, bit2 in enumerate(self.grdstates):
                if i == j:
                    page.td('&mdash;', class_='ignore')
                    continue
                count = len([t for t in self.transitions[bit] if
                             t[1] == bit2])
                if count:
                    page.td(str(count))
                else:
                    page.td('&mdash;')
            page.th(str(len(self.transitions[bit])))
            page.tr.close()
        page.tbody.close()
        page.table.close()
        page.button(
            'Export to CSV', class_='btn btn-outline-secondary btn-table mt-2',
            **{'data-table-id': id_, 'data-filename': '%s.csv' % id_})
        page.div.close()  # col-md-12
        page.div.close()  # row

        # summarise each state
        page.div(class_='row')
        page.div(class_='col-md-12')
        page.h2('State details', class_='mt-4 mb-2')
        page.div(id='accordion')
        for i, bit in enumerate(self.grdstates):
            name = self.grdstates[bit].strip('*')
            page.div(class_='card border-info mb-1 shadow-sm', id=str(bit))
            # heading
            page.div(class_='card-header text-white bg-info')
            page.a('%s [%d]' % (name, bit), href='#collapse-%d' % bit,
                   class_='card-link cis-link collapsed',
                   **{'data-toggle': 'collapse', 'aria-expandedt': 'false'})
            page.div.close()  # card-header

            # body
            page.div(id='collapse-%d' % bit, class_='collapse',
                     **{'data-parent': '#accordion'})
            page.div(class_='card-body')

            # print transitions
            headers = ['Transition GPS time', 'UTC time', 'Local time',
                       'Transition from', 'Exited to']
            data = []
            if self.ifo in ['H1', 'C1', 'P1']:
                localzone = tz.gettz('America/Los_Angeles')
            elif self.ifo in ['L1']:
                localzone = tz.gettz('America/Chicago')
            elif self.ifo in ['K1']:
                localzone = tz.gettz('Asia/Tokyo')
            else:
                localzone = tz.gettz('Europe/Berlin')
            for t, from_, to_ in self.transitions[bit]:
                t2 = Time(t, format='gps', scale='utc')
                tlocal = Time(
                    t2.datetime.replace(tzinfo=UTC).astimezone(localzone),
                    format='datetime', scale='utc')
                data.append((
                    t, t2.iso, tlocal.iso,
                    '%s [%d]' % (self.grdstates.get(from_, 'Unknown'), from_),
                    '%s [%d]' % (self.grdstates.get(to_, 'Unknown'), to_)))
            page.add(str(html.table(
                headers, data,
                id='%s-guardian-%s' % (self.ifo.lower(), str(bit)),
                caption="Transitions for %s %r state" % (self.node, name))))

            # print segments
            flag = get_segments(self.segmenttag % name, state.active,
                                query=False).copy()
            livetime = abs(flag.active)
            try:
                duty = abs(flag.active) / float(abs(flag.known)) * 100.
            except ZeroDivisionError:
                duty = 0
            page.p('This state was active for %.2f seconds (%.2f%%) during '
                   'the following segments:' % (livetime, duty))
            page.add(str(self.print_segments(flag)))
            page.div.close()  # card-body
            page.div.close()  # collapse
            page.div.close()  # card
        page.div.close()  # accordion
        page.div.close()  # col-md-12
        page.div.close()  # row

        return super(DataTab, self).write_state_html(state, plots=False,
                                                     pre=page)


register_tab(GuardianTab)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Definitions for range plots
"""

from math import pi
from heapq import nlargest
from itertools import combinations

import numpy

from matplotlib.ticker import (LogLocator, MaxNLocator)

import gwpy.astro
from gwpy.segments import (Segment, SegmentList)
from gwpy.timeseries import TimeSeries

from gwdetchar.plot import texify

from .registry import (get_plot, register_plot)
from .utils import hash
from ..data import (get_range_channel, get_range,
                    get_range_spectrogram, get_timeseries)
from ..segments import get_segments
from ..channels import split as split_channels

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- utils --------------------------------------------------------------------

RAN_DICT = {
        'sensemon_range': gwpy.astro.sensemon_range,
        'sensemon_range_psd': gwpy.astro.sensemon_range_psd,
        'inspiral_range': gwpy.astro.inspiral_range,
        'inspiral_range_psd': gwpy.astro.inspiral_range_psd,
        'burst_range': gwpy.astro.burst_range,
        'burst_range_psd': gwpy.astro.burst_range_spectrum,
}


def _get_params(keys, pargs, nchans=1):
    """Return a `dict` of `list` of plot arguments for every channel
    """
    params = {}
    for key in keys:
        try:
            value = pargs.pop(key)
        except KeyError:
            continue
        if not isinstance(value, (tuple, list)):
            value = [value] * nchans
        params[key] = value
    return params


# -- sensitive range ----------------------------------------------------------

class RangePlotMixin(object):
    data = 'spectrogram'
    _threadsafe = False
    defaults = {
        'snr': 8.0,
        'stride': 60.,
        'fftlength': 8,
        'overlap': 4,
        'fmin': 10,
    }

    def __init__(self, *args, **kwargs):
        super(RangePlotMixin, self).__init__(*args, **kwargs)
        self.fftparams = _get_params(
            ['stride', 'fftlength', 'overlap'],
            self.pargs, nchans=len(self.channels))
        self.rangeparams = _get_params(
            ['mass1', 'mass2', 'snr', 'energy', 'fmin', 'fmax', 'range_func'],
            self.pargs, nchans=len(self.channels))
        self.range_func = (get_range_spectrogram if
                           'spec' in self.type else get_range)

    def draw(self):
        """Read in all necessary data and generate a figure
        """
        keys = []
        # generate data
        for i, channel in enumerate(self.channels):
            fftkwargs = dict((key, self.fftparams[key][i]) for
                             key in self.fftparams if
                             self.fftparams[key][i] is not None)
            rangekwargs = dict((key, self.rangeparams[key][i]) for
                               key in self.rangeparams if
                               self.rangeparams[key][i] is not None)
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            # replace range_func arg with correct method
            if 'range_func' in rangekwargs.keys():
                rangekwargs['range_func'] = RAN_DICT[rangekwargs['range_func']]
            rlist = self.range_func(channel, valid, query=self.read,
                                    **fftkwargs, **rangekwargs)
            try:
                keys.append(str(rlist[0].channel))
            except IndexError:
                keys.append(get_range_channel(channel, **rangekwargs))

        # reset channel lists and generate plot
        channels = self.channels
        self.channels = keys
        out = super(RangePlotMixin, self).draw()
        self.channels = channels
        return out


class RangeDataPlot(RangePlotMixin, get_plot('timeseries')):
    type = 'range'
    defaults = get_plot('timeseries').defaults.copy()
    defaults.update(RangePlotMixin.defaults.copy())
    defaults.update({
        'ylabel': 'Sensitive distance [Mpc]',
    })


register_plot(RangeDataPlot)


class RangeDataHistogramPlot(RangePlotMixin, get_plot('histogram')):
    type = 'range-histogram'
    defaults = get_plot('histogram').defaults.copy()
    defaults.update(RangePlotMixin.defaults.copy())
    defaults.update({
        'xlabel': 'Sensitive distance [Mpc]',
    })


register_plot(RangeDataHistogramPlot)


class RangeSpectrogramDataPlot(RangePlotMixin, get_plot('spectrogram')):
    type = 'range-spectrogram'
    defaults = get_plot('spectrogram').defaults.copy()
    defaults.update(RangePlotMixin.defaults.copy())
    defaults.update({
        'cmap': 'inferno',
        'norm': 'linear',
    })


register_plot(RangeSpectrogramDataPlot)


class RangeSpectrumDataPlot(RangePlotMixin, get_plot('spectrum')):
    type = 'range-spectrum'
    data = 'spectrum'
    defaults = get_plot('spectrum').defaults.copy()
    defaults.update(RangePlotMixin.defaults.copy())
    defaults.update({
        'xlabel': 'Frequency [Hz]',
        'yscale': 'linear',
    })


register_plot(RangeSpectrumDataPlot)


class RangeCumulativeSpectrumDataPlot(RangePlotMixin, get_plot('spectrum')):
    type = 'cumulative-range-spectrum'
    data = 'spectrum'
    defaults = get_plot('spectrum').defaults.copy()
    defaults.update(RangePlotMixin.defaults.copy())
    defaults.update({
        'xlabel': 'Frequency [Hz]',
        'ylim': [0, 100],
        'ylabel': 'Cumulative fraction of range [%]',
        'yscale': 'linear',
        'yticks': [0, 20, 40, 60, 80, 100],
        'ytickmarks': [0, 20, 40, 60, 80, 100],
    })


register_plot(RangeCumulativeSpectrumDataPlot)


# -- time-volume --------------------------------------------------------------

class SimpleTimeVolumeDataPlot(get_plot('segments')):
    """Time-series of the time-volume searched by an interferometer
    """
    data = 'timeseries'
    type = 'time-volume'
    DRAW_PARAMS = get_plot('timeseries').DRAW_PARAMS
    defaults = get_plot('timeseries').defaults.copy()

    parse_plot_kwargs = get_plot('timeseries').parse_plot_kwargs

    def __init__(self, sources, *args, **kwargs):
        if isinstance(sources, str):
            sources = split_channels(sources)
        channels = sources[::2]
        flags = sources[1::2]
        get_plot('timeseries').__init__(self, channels, *args, **kwargs)
        self._allflags = []
        self.flags = flags

    @classmethod
    def from_ini(cls, *args, **kwargs):
        return get_plot('timeseries').from_ini(cls, *args, **kwargs)

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            self._pid = hash("".join(map(str, self.channels+self.flags)))
            return self.pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    @pid.deleter
    def pid(self):
        del self._pid

    @staticmethod
    def calculate_time_volume(segments, range):
        try:
            ts = TimeSeries(numpy.zeros(range.size), xindex=range.times,
                            unit='s')
        except IndexError:
            ts = TimeSeries(numpy.zeros(range.size), unit='s',
                            x0=range.x0, dx=range.dx)
        dx = range.dx.value

        # override range units
        range.override_unit('Mpc')

        # use float, not LIGOTimeGPS for speed
        segments = type(segments)([type(s)(float(s[0]), float(s[1])) for
                                  s in segments])

        def livetime_(t):
            return float(abs(SegmentList([Segment(t, t+dx)]) & segments))

        livetime = numpy.vectorize(livetime_, otypes=[float])
        ts[:] = livetime(ts.times.value) * ts.unit
        return (4/3. * pi * ts * range ** 3).to('Mpc^3 kyr')

    def combined_time_volume(self, allsegments, allranges):
        try:
            combined_range = TimeSeries(numpy.zeros(allranges[0].size),
                                        xindex=allranges[0].times, unit='Mpc')
        except IndexError:
            combined_range = TimeSeries(
                numpy.zeros(allranges[0].size), unit='Mpc',
                x0=allranges[0].x0, dx=allranges[0].dx)

        # get coincident observing segments
        pairs = list(combinations(allsegments, 2))
        coincident = SegmentList()
        for pair in pairs:
            coincident.extend(pair[0] & pair[1])
        coincident = coincident.coalesce()

        # get effective network range
        values = [r.value for r in allranges]
        values = [min(nlargest(2, x)) for x in zip(*values)]
        size = min([r.size for r in allranges])
        combined_range[:size] = values * combined_range.unit

        # compute time-volume
        return self.calculate_time_volume(coincident, combined_range)

    def draw(self, outputfile=None):
        """Generate the figure for this plot
        """
        plot = self.init_plot()
        ax = plot.axes[0]

        # get plotting arguments
        cumulative = self.pargs.pop('cumulative', False)
        plotargs = self.parse_plot_kwargs()
        legendargs = self.parse_legend_kwargs()

        # set ylabel
        if cumulative:
            self.pargs.setdefault(
                'ylabel',
                'Cumulative time-volume [Mpc$^3$ kyr]',
            )
        else:
            self.pargs.setdefault(
                'ylabel',
                'Time-volume [Mpc$^3$ kyr]',
            )

        # get data
        allsegs, allranges = ([], [])
        for channel, flag, pargs in zip(self.channels, self.flags, plotargs):
            pad = 0
            if self.state and not self.all_data:
                valid = self.state.active
                pad = numpy.nan
            elif channel.sample_rate.value:
                valid = SegmentList([self.span.protract(
                    1/channel.sample_rate.value)])
            else:
                valid = SegmentList([self.span])
            data = get_timeseries(
                channel, valid, query=False).join(gap='pad', pad=pad)
            if not data.unit or data.unit.to_string() in ['', 'undef']:
                data.override_unit('Mpc')
            segments = get_segments(flag, valid, query=False)
            timevolume = self.calculate_time_volume(segments.active, data)
            if cumulative:
                ax.plot(timevolume.cumsum(), **pargs)
            else:
                ax.plot(timevolume, **pargs)
            allsegs.append(segments.active)
            allranges.append(data)

        # estimate combined time-volume
        if self.all_data and len(self.channels) > 1:
            pargs = plotargs[-1]
            pargs['color'] = '#000000'
            pargs['label'] = 'Combined'
            pargs['linestyle'] = '--'
            combined_timevolume = self.combined_time_volume(
                allsegs, allranges)
            if cumulative:
                ax.plot(combined_timevolume.cumsum(), **pargs)
            else:
                ax.plot(combined_timevolume, **pargs)

        # add horizontal lines to add
        for yval in self.pargs.get('hline', []):
            try:
                yval = float(yval)
            except ValueError:
                continue
            else:
                ax.plot([self.start, self.end], [yval, yval],
                        linestyle='--', color='red')

        # customise plot
        self.apply_parameters(ax, **self.pargs)
        if (len(self.channels) > 1 or plotargs[0].get('label', None) in
                [texify(str(self.channels[0])), None]):
            ax.legend(**legendargs)

        # add extra axes and finalise
        self.add_state_segments(ax)
        self.add_future_shade()
        if ax.get_yscale() == 'log':
            ax.yaxis.set_major_locator(LogLocator())
        else:
            ax.yaxis.set_major_locator(MaxNLocator(8))
        ticks = ax.get_yticks()
        ax.yaxis.set_ticklabels(ticks)
        return self.finalize(outputfile=outputfile)


register_plot(SimpleTimeVolumeDataPlot)


class GWpyTimeVolumeDataPlot(RangePlotMixin, SimpleTimeVolumeDataPlot):
    """TimeVolumeDataPlot where the range is calculated on-the-fly
    """
    type = 'strain-time-volume'
    _threadsafe = False


register_plot(GWpyTimeVolumeDataPlot)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for gwsumm.html.static
"""

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

from collections import OrderedDict

from .. import static


# test simple utils

def test_get_css():
    css = static.get_css()
    assert isinstance(css, OrderedDict)
    # test dict keys
    assert list(css.keys()) == [
        'font-awesome',
        'font-awesome-solid',
        'gwbootstrap',
    ]
    # test list of files
    css_files = list(x.split('/')[-1] for x in css.values())
    assert css_files == [
        'fontawesome.min.css',
        'solid.min.css',
        'gwbootstrap.min.css',
    ]


def test_get_js():
    js = static.get_js()
    assert isinstance(js, OrderedDict)
    # test dict keys
    assert list(js.keys()) == [
        'jquery',
        'jquery-ui',
        'moment',
        'bootstrap',
        'fancybox',
        'datepicker',
        'gwbootstrap',
    ]
    # test list of files
    js_files = list(x.split('/')[-1] for x in js.values())
    assert js_files == [
        'jquery-3.5.1.min.js',
        'jquery-ui.min.js',
        'moment.min.js',
        'bootstrap.bundle.min.js',
        'jquery.fancybox.min.js',
        'bootstrap-datepicker.min.js',
        'gwbootstrap-extra.min.js',
    ]
